@@ -7,8 +7,8 @@
 #include <sys/mount.h>
 
 #include <filesystem>
-#include <map>
 #include <optional>
+#include <unordered_map>
 #include <vector>
 
 #include <sys/resource.h>
@@ -123,6 +123,7 @@ struct config
         std::optional<std::vector<id_mapping_t>> gid_mappings;
         std::optional<std::vector<std::filesystem::path>> masked_paths;
         std::optional<std::vector<std::filesystem::path>> readonly_paths;
+        unsigned int rootfs_propagation{ 0 };
     };
 
     std::optional<linux_t> linux;
@@ -133,7 +134,7 @@ struct config
         {
             std::filesystem::path path;
             std::optional<std::vector<std::string>> args;
-            std::optional<std::map<std::string, std::string>> env;
+            std::optional<std::unordered_map<std::string, std::string>> env;
             std::optional<int> timeout;
         };
 
@@ -169,6 +170,8 @@ struct config
     };
 
     root_t root;
+
+    std::optional<std::unordered_map<std::string, std::string>> annotations;
 };
 
 } // namespace linyaps_box
